"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit warnings properties report function. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Warnings.warnings import get_warnings
from duHast.Revit.Warnings.Objects.warnings_storage import RevitWarning
from duHast.Utilities.Objects import result as res
from duHast.Revit.Warnings.Reporting.warnings_report_header import (
    REPORT_WARNINGS_HEADER,
)
from duHast.Utilities.files_csv import write_report_data_as_csv


def get_warnings_report_data(doc, revit_file_name):
    """
    Gets warnings data to be written to report file.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: The file hostname, which is added to data returned.
    :type revit_file_path: str
    :return: list of list of sheet properties.
    :rtype: list of list of str
    """

    # get all warnings
    data = {}
    warnings = get_warnings(doc)
    for warning in warnings:
        try:
            failure_id_guid = warning.GetFailureDefinitionId().Guid
            failure_description = warning.GetDescriptionText()
            failing_ids = warning.GetFailingElements()
            failing_ids_as_integer = []
            for fail_id in failing_ids:
                failing_ids_as_integer.append(fail_id.IntegerValue)

            warning_stored = RevitWarning(
                file_name=revit_file_name,
                id=failure_id_guid,
                description=failure_description,
                element_ids=failing_ids_as_integer,
            )
            if warning_stored.id in data:
                data[warning_stored.id].append(warning_stored)
            else:
                data[warning_stored.id] = [warning_stored]

        except Exception as e:
            # store the exception
            warning_exception = ResourceWarning(
                file_name=revit_file_name,
                id=-1,
                description="failed to retrieve warning with exception: {}".format(e),
                element_ids=[],
            )
            if warning_stored.id in data:
                data[warning_stored.id].append(warning_exception)
            else:
                data[warning_stored.id] = [warning_exception]

    return data


def convert_warnings_data_to_list(warnings_data):
    """
    Converts a list of dictionaries of view properties names and values to a list of properties only.

    :param view_data: List of dictionaries representing view properties
    :type view_data: [{}]
    :return: A list of lists of view property values.
    :rtype: [[str]]
    """

    data = []
    for key, value in warnings_data.items():
        number_elements_affected = 0
        dummy = None
        for warning_data_instance in value:
            number_elements_affected = number_elements_affected + len(
                warning_data_instance.element_ids
            )
            dummy = warning_data_instance
        # build a list with entries:
        # file name, warning GUID, number of warnings belonging to warnings GUID, warnings description, number of elements involved overall
        data.append(
            [
                dummy.file_name,
                dummy.date,
                dummy.time,
                dummy.id,
                len(warnings_data[key]),
                dummy.description,
                number_elements_affected,
            ]
        )
    return data


def write_warnings_data(file_name, data):
    """
    Writes to file all warnings properties.

    file type: csv

    :param file_name: The fully qualified file path of the report file.
    :type file_name: str
    :param current_file_name: The current revit file name which will be appended to data in the report.
    :type current_file_name: str
    :return:
        Result class instance.
        - .status True if data was written successfully. Otherwise False.
        - .message will contain write status.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        data_converted = convert_warnings_data_to_list(warnings_data=data)
        write_report_data_as_csv(file_name, REPORT_WARNINGS_HEADER, data_converted)
        return_value.update_sep(
            True, "Successfully wrote data file at {}".format(file_name)
        )
    except Exception as e:
        return_value.update_sep(False, str(e))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Duplicate mark warnings solver class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common import parameter_get_utils as rParaGet
from duHast.Revit.Common import parameter_set_utils as rParaSet
from duHast.Utilities.Objects import result as res


# import Autodesk
import Autodesk.Revit.DB as rdb
from duHast.Utilities.Objects import base


class RevitWarningsSolverDuplicateMark(base.Base):

    def __init__(self, filter_func, filter_values=[]):
        """
        Constructor: this solver takes two arguments: a filter function and a list of values to filter by

        :param filter_func: A function to filter elements in warnings by
        :type filter_func: func(document, elementId, list of filter values)
        :param filter_values: A list of filter values, defaults to []
        :type filter_values: list, optional
        """

        # ini super class to allow multi inheritance in children!
        super(RevitWarningsSolverDuplicateMark, self).__init__()

        self.filter = filter_func
        self.filter_values = filter_values
        self.filter_name = "Duplicate mark value."

    # --------------------------- duplicate mark guid ---------------------------
    #: guid identifying this specific warning
    GUID = "6e1efefe-c8e0-483d-8482-150b9f1da21a"

    IGNORED_WARNINGS = ["Type Mark"]

    def solve_warnings(self, doc, warnings):
        """
        Solver setting element mark to nothing, provided it passes the filter.

        :param doc: Current Revit model document.
        :type doc: Autodesk.Revit.DB.Document
        :param warnings: List of warnings to be solved.
        :type warnings: Autodesk.Revit.DB.FailureMessage

        :return:
            Result class instance.

            - .status True if all duplicate mark warnings could be solved. Otherwise False.
            - .message will contain stats in format parameter value set to ''

        :rtype: :class:`.Result`
        """

        return_value = res.Result()
        if len(warnings) > 0:
            for warning in warnings:
                # Flag to indicate if we should continue the outer loop
                should_continue_outer = False
                # check for any duplicate Type Mark warnings...which are not to be addressed!
                for ignore in self.IGNORED_WARNINGS:
                    if ignore in warning.GetDescriptionText():
                        element_ids = warning.GetFailingElements()
                        for el_id in element_ids:
                            element = doc.GetElement(el_id)
                            return_value.update_sep(
                                True,
                                "{} Warning of type: duplicate {}. {} will be ignored.".format(
                                    self.filter_name,
                                    rdb.Element.Name.GetValue(element),
                                    ignore,
                                ),
                            )
                        should_continue_outer = True
                        break  # Break out of the inner loop
                if should_continue_outer:
                    continue  # Continue the outer loop
                element_ids = warning.GetFailingElements()
                for el_id in element_ids:
                    element = doc.GetElement(el_id)
                    # check whether element passes filter
                    if self.filter(doc, el_id, self.filter_values):

                        try:
                            p_value = rParaGet.get_built_in_parameter_value(
                                element, rdb.BuiltInParameter.ALL_MODEL_MARK
                            )
                            if p_value != None:
                                result = rParaSet.set_built_in_parameter_value(
                                    doc,
                                    element,
                                    rdb.BuiltInParameter.ALL_MODEL_MARK,
                                    "",
                                )
                                return_value.update(result)
                            else:
                                return_value.update_sep(
                                    True,
                                    "{}:  Element has no mark value: {}".format(
                                        self.filter_name,
                                        rdb.Element.Name.GetValue(element),
                                    ),
                                )
                        except Exception as e:
                            return_value.update_sep(
                                False,
                                "{}: Failed to solve warning duplicate mark with exception: {}".format(
                                    self.filter_name, e
                                ),
                            )
                    else:
                        return_value.update_sep(
                            True,
                            "{}: Element removed by filter: {}".format(
                                self.filter_name, rdb.Element.Name.GetValue(element)
                            ),
                        )
        else:
            return_value.update_sep(
                True,
                "{}: No warnings of type: duplicate mark in model.".format(
                    self.filter_name
                ),
            )
        return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a purge unused function using Autodesk's own eTransmit plug in.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Based on the building coder article:
https://thebuildingcoder.typepad.com/blog/2022/03/purge-unused-and-the-autodesk-camel.html
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import clr

# import Autodesk.Revit.DB as rdb

from duHast.Utilities.Objects import result as res
from duHast.Utilities.Objects.timer import Timer


def _purge(doc, dll_path):
    """
    Purges the document using the purge unused functionality of the eTransmit tool provided by Autodesk.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param dll_path: Fully qualified file path to the revit version specific eTransmitForRevitDB.dll installed by eTransmit
    :type dll_path: str

    :return: True purge was successful, otherwise False.
    :rtype: bool
    """

    clr.AddReferenceToFileAndPath(dll_path)
    # import the eTransmit name space which includes
    # purge unused functionality
    from eTransmitForRevitDB import eTransmitUpgradeOMatic, UpgradeFailureType

    # purge the document
    e_transmit_upgrade_o_matic_thing = eTransmitUpgradeOMatic(doc.Application)
    result = e_transmit_upgrade_o_matic_thing.purgeUnused(doc)
    return result == UpgradeFailureType.UpgradeSucceeded


# -------------------------------------------- Purge Unused using eTransmit for Revit 2019 -------------------------------------


def _purge_unused_2019(doc):
    """
    Purges the document, revit version 2019, using the purge unused functionality of the eTransmit tool provided by Autodesk

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: True purge was successful, otherwise False.
    :rtype: bool
    """

    # path to e-Transmit dll for Revit 2019
    e_transmit_file_path = (
        r"C:\Program Files\Autodesk\eTransmit for Revit 2019\eTransmitForRevitDB.dll"
    )
    value = _purge(doc, e_transmit_file_path)
    return value


# -------------------------------------------- Purge Unused using eTransmit for Revit 2020 -------------------------------------


def _purge_unused_2020(doc):
    """
    Purges the document, revit version 2020, using the purge unused functionality of the eTransmit tool provided by Autodesk

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: True purge was successful, otherwise False.
    :rtype: bool
    """

    # path to e-Transmit dll for Revit 2019
    e_transmit_file_path = (
        r"C:\Program Files\Autodesk\eTransmit for Revit 2020\eTransmitForRevitDB.dll"
    )
    value = _purge(doc, e_transmit_file_path)
    return value


# -------------------------------------------- Purge Unused using eTransmit for Revit 2021 -------------------------------------


def _purge_unused_2021(doc):
    """
    Purges the document, revit version 2021, using the purge unused functionality of the eTransmit tool provided by Autodesk

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: True purge was successful, otherwise False.
    :rtype: bool
    """

    # path to e-Transmit dll for Revit 2019
    e_transmit_file_path = (
        r"C:\Program Files\Autodesk\eTransmit for Revit 2021\eTransmitForRevitDB.dll"
    )
    value = _purge(doc, e_transmit_file_path)
    return value


# -------------------------------------------- Purge Unused using eTransmit for Revit 2021 -------------------------------------


def _purge_unused_2022(doc):
    """
    Purges the document, revit version 2022, using the purge unused functionality of the eTransmit tool provided by Autodesk

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: True purge was successful, otherwise False.
    :rtype: bool
    """

    # path to e-Transmit dll for Revit 2019
    e_transmit_file_path = (
        r"C:\Program Files\Autodesk\eTransmit for Revit 2022\eTransmitForRevitDB.dll"
    )
    value = _purge(doc, e_transmit_file_path)
    return value


# -------------------------------------------- Purge Unused using eTransmit for Revit 2021 -------------------------------------


def purge_unused_e_transmit(doc):
    """
    Purges the document using the purge unused functionality of the eTransmit tool provided by Autodesk.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return:
        Result class instance.

        - Purge status returned in result.status. False if an exception occurred, purge command returned False, unsupported Revit version, otherwise True.
        - result.message will contain the purge status.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    result_value = res.Result()
    t_overall = Timer()
    t_overall.start()
    # get the revit version:
    revit_version = doc.Application.VersionNumber
    try:
        if revit_version == "2019":
            result_value.status = _purge_unused_2019(doc)
        elif revit_version == "2020":
            result_value.status = _purge_unused_2020(doc)
        elif revit_version == "2021":
            result_value.status = _purge_unused_2021(doc)
        elif revit_version == "2022":
            result_value.status = _purge_unused_2022(doc)
        else:
            # this is a non supported revit version!
            raise ValueError(
                "Revit version: {} is currently not supported by purge unused!".format(
                    revit_version
                )
            )
        # update messaging
        if result_value.status == True:
            result_value.message = "Successfully purged model!"
        else:
            result_value.message = "Failed to purge model!"
    except Exception as e:
        result_value.update_sep(
            False, "Terminated purge unused actions with exception: {}".format(e)
        )
    result_value.append_message("purge duration: {}".format(t_overall.stop()))
    return result_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit Geometry to data geometry conversion helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Data.Objects.Properties.Geometry import geometry_polygon as dGeometryPoly
from duHast.Revit.Common.Geometry import geometry as rGeo
from collections import namedtuple
from duHast.Revit.Common.Geometry.points import get_point_as_doubles, flatten_xyz_point

from Autodesk.Revit.DB import XYZ


def convert_xyz_in_data_geometry_polygons(doc, dgObject):
    """
    Converts Revit XYZ objects stored in a data geometry object into groups of doubles for inner and outer loops\
        and stores them in new data geometry object. It also populates translation and rotation matrix data of\
            coordinate system information.
    :param doc: _description_
    :type doc: _type_
    :param dgObject: A data geometry object.
    :type dgObject: :class:`.DataGeometryPolygon`
    :return: A data geometry object.
    :rtype: :class:`.DataGeometryPolygon`
    """

    data_geometry = dGeometryPoly.DataPolygon()
    outer_loop = []
    for xyz_point in dgObject.outer_loop:
        point_as_double = get_point_as_doubles(xyz_point)
        outer_loop.append(point_as_double)
    inner_loops = []
    for inner_loop in dgObject.inner_loops:
        inner_loop_points = []
        for xyz_point in inner_loop:
            point_as_double = get_point_as_doubles(xyz_point)
            inner_loop_points.append(point_as_double)
        inner_loops.append(inner_loop_points)
    data_geometry.outer_loop = outer_loop
    data_geometry.inner_loops = inner_loops
    # add coordinate system translation and rotation data
    (
        data_geometry.rotation_coord,
        data_geometry.translation_coord,
    ) = rGeo.get_coordinate_system_translation_and_rotation(doc)
    return data_geometry


def convert_solid_to_flattened_2d_points(solid):
    """
    Converts a solid into a 2D polygon by projecting it onto a plane.( Removes Z values...)
    First nested list is the outer loop, any other following lists describe holes within the area of the polygon defined be points in first list.
    Arcs, circles will be tessellated to polygons.
    :param solid: A solid.
    :type solid: Autodesk.Revit.DB.Solid
    :return: A list of data geometry instances.
    :rtype: list of :class:`.DataGeometryPolygon`
    """

    """
    sample for a sold with multiple sketches:
    [
        [
            [external poly line],[hole],[hole]
        ],
        [
            [external poly line] # without any holes
        ]
    ]
    sort faces into groups by volume:
    This may be required because a solid can be made up of multiple volumes (extrusion, sweeps etc)
    Turns out a solid returns a single face for multiple sketches. In order to work out whether these are multiple non overlapping polygons I will need to check
    whether a point from one polygon is within the other if so it may represents a hole or an island within a hole...to avoid misreading an island for a whole I will need to sort the faces by area
    and check from largest down to smallest.
    Also poly lines send back will always only represent: first list: exterior boundary as polygon any follow list is a hole within the polygon. Any islands in those holes will get their own top level representation
    i.e. no further list nesting!
    Within the faces groups: identify faces which are horizontal: its normal is facing up or down
    select the face with the lower Z coordinates and
    group all edges of the above face which form a closed loop (first loop of edges to describe the extend of that face, any secondary loops define holes in face)
    - > sort all edges by their connections (need to be connected by a point) so they describe a loop <- seems to be ok as revit provides them
    extract points of edges
    """

    ceilingGeos = []
    # sort faces by size
    sortedBySizeFaces = rGeo.get_faces_sorted_by_area_from_solid(solid)
    # get all faces which are horizontal only
    horizontalFaces = rGeo.get_unique_horizontal_faces(sortedBySizeFaces)
    # loop of all horizontal faces and extract loops
    for hf in horizontalFaces:
        edgeLoops = rGeo.convert_edge_arrays_into_list_of_points(hf.EdgeLoops)
        # convert in UV coordinates
        edgeLoopsFlattened = rGeo.flatten_xyz_point_list_of_lists(edgeLoops)
        # set up a named tuple to store data in it
        uvLoops = []
        uvLoop = namedtuple("uvLoop", "loop area id threeDPoly")
        counter = 0
        for edgeLoopFlat in edgeLoopsFlattened:
            areaLoop = rGeo.get_signed_polygon_area(edgeLoopFlat)
            uvTuple = uvLoop(edgeLoopFlat, abs(areaLoop), counter, edgeLoops[counter])
            uvLoops.append(uvTuple)
            counter += 1
        uvLoops = sorted(uvLoops, key=lambda x: x.area, reverse=True)
        # sort loops into exterior and hole loops
        loopDic = rGeo.build_loops_dictionary(uvLoops)
        for key in loopDic:
            dataGeometry = dGeometryPoly.DataPolygon()
            keyList = []
            # find matching loop by id
            for x in uvLoops:
                if x.id == key:
                    keyList = x
                    break
            dataGeometry.outer_loop = keyList.threeDPoly
            if len(loopDic[key]) > 0:
                for hole in loopDic[key]:
                    dataGeometry.inner_loops.append(hole.threeDPoly)
            else:
                dataGeometry.inner_loops = []
            ceilingGeos.append(dataGeometry)
    return ceilingGeos


def convert_bounding_box_to_flattened_2d_points(bounding_box):
    """
    Converts a bounding box into a 2D polygon by projecting it onto a plane.( Removes Z values...)
    :param bounding_box: A bounding box.
    :type bounding_box: Autodesk.Revit.DB.BoundingBoxXYZ
    :return: A list of data geometry instances.
    :rtype: list of :class:`.DataGeometryPolygon`
    """
    # get points from bounding box and project them onto a plane ( Z value of bounding box min point)
    bounding_box_points = [
        bounding_box.Min,
        XYZ(bounding_box.Max.X, bounding_box.Min.Y, bounding_box.Min.Z),
        XYZ(bounding_box.Max.X, bounding_box.Max.Y, bounding_box.Min.Z),
        XYZ(bounding_box.Min.X, bounding_box.Max.Y, bounding_box.Min.Z),
    ]
    # set up data class object and store points in outer loop property
    dataGeometry = dGeometryPoly.DataPolygon()
    dataGeometry.outer_loop = bounding_box_points
    return dataGeometry



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions to get the IFCExportConfig in varies versions of Revit.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System
import sys

import Autodesk.Revit.DB as rdb


# -------------------------------------------- IFC EXPORT Revit 2020 -------------------------------------
# need to check for new features in revit 2020


def ifc_get_third_party_export_config_by_view_2020(ifc_version, ifc_settings):
    """
    Function returning an IFC export configuration for Revit 2020 using the open source third party IFC exporter plug in supported by AutoDesk

    This configuration allows export by view. If ifc_version is None, IFCVersion.Default will be used.

    :param ifc_version: The ifc version used for the export (2x3 etc...).
    :type ifc_version: Autodesk.Revit.DB.IFCVersion
    :return: An IFCExportConfig object.
    :rtype: BIM.IFC.Export.UI.IFCExportConfiguration
    """

    # load version specific assemblies
    ifc_third_party_folder_path_ = r"C:\ProgramData\Autodesk\ApplicationPlugins\IFC 2020.bundle\Contents\2020\IFCExportUIOverride.dll"
    clr.AddReferenceToFileAndPath(ifc_third_party_folder_path_)
    # import the BIM namespace which includes
    # IFCExportConfiguration and IFCExportConfigurationMaps classes
    from BIM.IFC.Export.UI import IFCExportConfiguration

    # set up configuration
    ifc_export_config = IFCExportConfiguration.CreateDefaultConfiguration()
    ifc_export_config.Name = "DefaultIFCByViewSetup"
    # set up IFC version
    if ifc_version is None or ifc_version == "":
        ifc_export_config.IFCVersion = rdb.IFCVersion.Default
    else:
        ifc_export_config.IFCVersion = ifc_version

    ifc_export_config.SpaceBoundaries = 1
    ifc_export_config.ActivePhaseId = rdb.ElementId.InvalidElementId
    ifc_export_config.ExportBaseQuantities = True
    ifc_export_config.SplitWallsAndColumns = True
    ifc_export_config.VisibleElementsOfCurrentView = True  # by view
    ifc_export_config.Use2DRoomBoundaryForVolume = False
    ifc_export_config.UseFamilyAndTypeNameForReference = True
    ifc_export_config.ExportInternalRevitPropertySets = True
    ifc_export_config.ExportIFCCommonPropertySets = True
    ifc_export_config.Export2DElements = False
    ifc_export_config.ExportPartsAsBuildingElements = True
    ifc_export_config.ExportBoundingBox = False
    ifc_export_config.ExportSolidModelRep = False
    ifc_export_config.ExportSchedulesAsPsets = False
    ifc_export_config.ExportUserDefinedPsets = False
    ifc_export_config.ExportUserDefinedPsetsFileName = ""
    ifc_export_config.ExportLinkedFiles = False
    ifc_export_config.IncludeSiteElevation = True
    ifc_export_config.UseActiveViewGeometry = False  # setting this value to True will slow down the IFC export considerably (sample: from 8min to 45min!)
    ifc_export_config.ExportSpecificSchedules = False
    ifc_export_config.TessellationLevelOfDetail = 0
    ifc_export_config.StoreIFCGUID = True
    ifc_export_config.ExportRoomsInView = (
        False  # might not work in 3D views if volumes are not computed???
    )
    # revit 2019.1
    ifc_export_config.UseOnlyTriangulation = False
    ifc_export_config.IncludeSteelElements = True
    ifc_export_config.COBieCompanyInfo = "Company Name"
    ifc_export_config.COBieProjectInfo = "Project Info"

    return ifc_export_config


# ifcVersion        which ifc version (2x3 etc...)
def ifc_get_third_party_export_config_by_model_2020(ifc_version, ifc_settings):
    """
    Function returning an IFC export configuration for Revit 2020 using the open source third party IFC exporter plug in supported by AutoDesk

    This configuration allows export the entire model. If ifc_version is None, IFCVersion.Default will be used.

    :param ifc_version: The ifc version used for the export.
    :type ifc_version: Autodesk.Revit.DB.IFCVersion
    :return: An IFCExportConfig object.
    :rtype: BIM.IFC.Export.UI.IFCExportConfiguration
    """

    # load version specific assemblies
    ifc_third_party_folder_path_ = r"C:\ProgramData\Autodesk\ApplicationPlugins\IFC 2020.bundle\Contents\2020\IFCExportUIOverride.dll"
    clr.AddReferenceToFileAndPath(ifc_third_party_folder_path_)
    # import the BIM namespace which includes
    # IFCExportConfiguration and IFCExportConfigurationMaps classes
    from BIM.IFC.Export.UI import IFCExportConfiguration

    # set up configuration
    ifc_export_config = IFCExportConfiguration.CreateDefaultConfiguration()
    ifc_export_config.Name = "DefaultIFCByModelSetup"

    # set up IFC version
    if ifc_version is None or ifc_version == "":
        ifc_export_config.IFCVersion = rdb.IFCVersion.Default
    else:
        ifc_export_config.IFCVersion = ifc_version

    ifc_export_config.SpaceBoundaries = 1
    ifc_export_config.ActivePhaseId = rdb.ElementId.InvalidElementId
    ifc_export_config.ExportBaseQuantities = True
    ifc_export_config.SplitWallsAndColumns = True
    ifc_export_config.VisibleElementsOfCurrentView = False  # by model
    ifc_export_config.Use2DRoomBoundaryForVolume = False
    ifc_export_config.UseFamilyAndTypeNameForReference = True
    ifc_export_config.ExportInternalRevitPropertySets = True
    ifc_export_config.ExportIFCCommonPropertySets = True
    ifc_export_config.Export2DElements = False
    ifc_export_config.ExportPartsAsBuildingElements = True
    ifc_export_config.ExportBoundingBox = False
    ifc_export_config.ExportSolidModelRep = False
    ifc_export_config.ExportSchedulesAsPsets = False
    ifc_export_config.ExportUserDefinedPsets = False
    ifc_export_config.ExportUserDefinedPsetsFileName = ""
    ifc_export_config.ExportLinkedFiles = False
    ifc_export_config.IncludeSiteElevation = True
    ifc_export_config.UseActiveViewGeometry = False  # by model
    ifc_export_config.ExportSpecificSchedules = False
    ifc_export_config.TessellationLevelOfDetail = 0
    ifc_export_config.StoreIFCGUID = True
    ifc_export_config.ExportRoomsInView = (
        False  # might not work in 3D views if volumes are not computed???
    )
    # revit 2019.1
    ifc_export_config.UseOnlyTriangulation = False
    ifc_export_config.IncludeSteelElements = True
    ifc_export_config.COBieCompanyInfo = "Company Name"
    ifc_export_config.COBieProjectInfo = "Project Info"

    return ifc_export_config



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to process Revit BatchProcessor log files.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# collection of tools used to process batch processor log files

# todo:
# write out current process IDs (empty text file with process id as file name)
#       script_util.GetSessionId()
# read out process ID from text file and find matching log files
# process log file:
#   - find files processed:
#            11/12/2020 18:01:40 : Processing Revit file (1 of 1) in Revit 2020 session.
#            11/12/2020 18:01:40 :
#            11/12/2020 18:01:40 : 	P:\something\FileName.rvt
#            ....
#            25/11/2020 09:37:34 : 	- Operation completed.
#   - check whether an exception occurred when processing any of the above files:
#       - file(s) not found
#           "WARNING: The following Revit Files do not exist"
#       - .net exception
#           "ERROR: An error occurred while executing the task script!"
#       - timed out occurred and revit process got killed
#           "WARNING: Timed-out"
#   - delete process id pointer file
#   - show user processing results (only when something went wrong(?))

import clr
import System

import glob
import time
import os
import json

# custom result class from common library
from duHast.Utilities.Objects import result as res

# library from common library
from duHast.Utilities import files_io as fileIO, files_get as fileGet

#: global variable controlling debug output
debug_mode_ = False

#: Message snippets which indicate processing of a file went bad
EXCEPTION_MESSAGES = [
    "ERROR: An error occurred while executing the task script! Operation",  # revit batch p message when script is buggy
    "WARNING: Timed-out",  # revit batch p message when revit times out
    "Exception: [Exception]",  # Revit Batch P message when an exception occurs
    "\t- \tMessage: An unrecoverable error has occurred.  The program will now be terminated.",  # Revit message when it crashes out
    "Script Exception:",  # custom script message...something the script was meant to do failed
]

# output...
def output(message=""):
    if debug_mode_:
        print(message)


def delete_log_data_files(directory_path):
    """
    Deletes all log marker files in a directory.

    :param directory_path: The directory path containing marker files to be deleted.
    :type directory_path: str

    :return: True if all files where deleted successfully, otherwise False.
    :rtype: bool
    """

    status = True
    # get files in directory
    files_to_delete = fileGet.get_files_with_filter(directory_path, ".txt")
    if len(files_to_delete) > 0:
        status_delete = True
        for file in files_to_delete:
            status_delete = status_delete and fileIO.file_delete(file)
    return status


def adjust_session_id_for_file_name(id):
    """
    Removes chevrons and replace colons with underscores in session id supplied by revit batch processor so it\
        can be used in a file name.

    :param id: Session id supplied by revit batch processor.
    :type id: str

    :return: Re-formatted session id.
    :rtype: str
    """

    # remove colons
    session_id_changed = id.replace(":", "_")
    # remove chevrons
    session_id_changed = session_id_changed[1:-1]
    return session_id_changed


def adjust_session_id_file_name_back(file_name_id):
    """
    Re-introduces chevrons and replaces underscores with colons to match session Id format used in batch processor to a\
        file name using a batch processor supplied id.

    :param file_name_id: A file name containing a session id with all illegal characters replaced.
    :type file_name_id: str

    :return: A session id.
    :rtype: str
    """

    # re-instate colons
    session_id_changed = file_name_id.replace("_", ":")
    # remove chevrons
    session_id_changed = "<" + session_id_changed + ">"
    return session_id_changed


def write_session_id_marker_file(folder_path, session_id):
    """
    Writes out an empty marker file in given directory.

    - File is of type .txt
    - File name is the batch processor session_id used to identify the log file belonging to this process.

    :param folder_path: Directory of where the file will be written to.
    :type folder_path: str
    :param session_id: Session id supplied by revit batch processor.
    :type session_id: str

    :return: True if marker file was written successfully, otherwise False.
    :rtype: bool
    """

    file_name = os.path.join(folder_path, str(session_id) + ".txt")
    status = True
    try:
        f = open(file_name, "w")
        f.close()
    except:
        status = False
    return status


def get_current_session_ids(folder_path, delete_marker_files=True):
    """
    Returns file names of all text files in a given directory representing session Ids.

    Files will be deleted immediately after reading

    :param folder_path: Directory of where test files are located
    :type folder_path: str

    :return: A list of ids in string format.
    :rtype: [str]
    """
    ids = []
    file_list = glob.glob(folder_path + "\\*" + ".txt")
    # delete marker files if required
    if(delete_marker_files == True):
        for fd in file_list:
            fileIO.file_delete(fd)
        result_delete = True
        for fd in file_list:
            if debug_mode_ == False:
                result_delete = result_delete & fileIO.file_delete(fd)
        if not result_delete:
            output("Failed to delete a marker file!")
    for f in file_list:
        ids.append(
            adjust_session_id_file_name_back(fileIO.get_file_name_without_ext(f))
        )
    return ids


def get_log_files(list_of_session_ids):
    """
    Returns a list of fully qualified filepath to log files matching the provided session Ids.

    :param list_of_session_ids: List of session ids.
    :type list_of_session_ids: [str]

    :return: List of fully qualified file path.
    :rtype: [str]
    """

    # save the current file in epoch
    time_now = time.time()
    file_list = glob.glob(
        os.path.join(os.getenv("LOCALAPPDATA"), "BatchRvt") + "\\*" + ".log"
    )
    log_files = []
    if len(file_list) > 0:
        for l in file_list:
            # check whether file is older than 24h
            file_time = os.path.getmtime(l)
            # 24 hr are 86400 seconds
            time_out = 86400
            if debug_mode_ == True:
                time_out = 8640000
            if time_now - file_time < time_out:
                # read the first two rows of the file to get the id
                id_string = get_session_id_from_log_file(l)
                for id_to_match in list_of_session_ids:
                    if id_to_match == id_string:
                        log_files.append(l)
            else:
                output("File is to old: " + str(l))
    return log_files


def get_session_id_from_log_file(file_path):
    """
    Reads the first two rows of a log file to get the session Id used.

    :param file_path: Fully qualified file path to log file
    :type file_path: str

    :return: The session id, or if not not a log file: an empty string.
    :rtype: str
    """

    row_counter = 0
    retrieved_id = ""
    for line in open(file_path, "r"):
        # read row 2 only
        if row_counter == 1:
            data = json.loads(line)
            message = get_message_from_json(data)
            retrieved_id = get_id_from_row(message)
            break
        row_counter += 1
    return retrieved_id


def get_id_from_row(row):
    """
    Extracts the session Id from logfile row.

    sample:
    {"date":{"local":"17/12/2020","utc":"17/12/2020"},"time":{"local":"16:49:27","utc":"05:49:27"},"sessionId":"235e2180-dc33-4d61-8773-1005a59344c0","message":{"msgId":"","message":"Session ID: <2020-12-17T05:49:27.559Z>"}}

    Note: There are two session id fields in thi string!

    TODO: Do some error handling!

    :param row: A json formatted string with the session id in chevrons.
    :type row: str

    :return: The session id.
    :rtype: str
    """

    first = "<"
    last = ">"
    t = get_text_between(row, first, last)
    return first + t + last


def get_text_between(text, first, last):
    """
    Returns text in between characters

    :param text: Text to parse
    :type text: str
    :param first: String indicating start
    :type first: str
    :param last: String indicating end
    :type last: str

    :return: string in between start and end.
    :rtype: str
    """

    start = text.index(first) + len(first)
    end = text.index(last, start)
    return text[start:end]


def get_message_from_json(data):
    """
    Returns the outer message string from json formatted message field in log file.

    sample:
    {"date":{"local":"17/12/2020","utc":"17/12/2020"},"time":{"local":"16:49:27","utc":"05:49:27"},"sessionId":"235e2180-dc33-4d61-8773-1005a59344c0","message":{"msgId":"","message":"Session ID: <2020-12-17T05:49:27.559Z>"}}

    This includes leading tab characters

    :param data: json formatted row of logfile
    :type data: str

    :return: The message string
    :rtype: str
    """
    outer_message = data["message"]
    return outer_message["message"]


def process_log_file(file_path):
    """
    Process revit batch processor session log file.

    - find Revit files processed:
    - check whether an exception occurred when processing any of the above files.

    :param file_path: Fully qualified file path to json formatted log file
    :type file_path: str

    :return: returns list of arrays in format:
        [[processed Revit file name, status of processing (true or false), message]]
    :rtype: [[str]]
    """

    files_process_status = []
    # get all files processed
    try:
        files_processed = get_files_processed(file_path)
        # check whether any file not founds came back
        try:
            files_not_found = get_files_not_found(files_processed)
            try:
                # filter files_processed by files not found
                files_to_check = filter_files_not_found(
                    files_processed, files_not_found
                )
                # check for exceptions during file processing
                for file_to_check in files_to_check:
                    try:
                        status, message = get_process_status(file_to_check, file_path)
                    except Exception as e:
                        output("GetProcessStatus: " + str(e))
                    dummy = [file_to_check, status, message]
                    files_process_status.append(dummy)
                # add files not found
                for f in files_not_found:
                    dummy = [f[0], False, ["File not found"]]
                    files_process_status.append(dummy)
            except Exception as e:
                output("FileToCheck: " + str(e))
        except Exception as e:
            output("GetFilesNotFound: " + str(e))
    except Exception as e:
        output("GetFilesProcessed: " + str(e))
    return files_process_status


# filtering files not found from overall file list
#
# filesProcessed: list of arrays, first entry in array is fully qualified file path
# filesNotFound: list of fully qualified file path
# returns a list of fully qualified file path (of files marked as found)
def filter_files_not_found(files_processed, files_not_found):
    filtered_list = []
    for file_name, status in files_processed:
        flag = False
        for fn, fnStatus in files_not_found:
            if fn == file_name:
                flag = True
                break
        if not flag:
            filtered_list.append(file_name)
    return filtered_list


def get_process_status(file_to_check, log_file_path):
    """
    Reads a log file and checks whether any exception occurred when processing a specific revit file.

    :param file_to_check: Fully qualified file path of Revit file which was processed
    :type file_to_check: str
    :param log_file_path: The fully qualified batch processor session log file path.
    :type log_file_path: str

    :return: A process status and a message.

        - process status: True if no exception occurred during revit file processing, otherwise false
        - message: the exception message recorded in the log file.

    :rtype: bool, str
    """

    message = ["Found no match in log file"]
    # flag indicating whether we managed to retrieve processing data for a file
    found_match = False
    json_data = read_log_file(log_file_path)
    # get data block showing how each file was processed
    unformatted_revit_file_process_messages = get_log_blocks(
        json_data,
        "\t- Processing file (",
        ["\t- Task script operation completed.", "\t- Operation aborted."],
        True,
    )
    process_status = True
    # loop over messages in this block and check for time out, and exception messages
    for message_block in unformatted_revit_file_process_messages:
        # check if right file the file name
        # todo
        file_name = get_file_name_from_data_block(message_block)
        if file_name == file_to_check:
            output(
                "file to check: "
                + str(file_to_check + "     file found: " + file_name)
                + "    is match "
                + str(file_name == file_to_check)
            )
            found_match = True
            # flag to show whether logs show any issues
            found_problem = False
            for m in message_block:
                output(m)
                # check for exceptions
                for exception_message in EXCEPTION_MESSAGES:
                    if exception_message in m:
                        process_status = False
                        found_problem = True
                        message = [m.strip()]
                        break
            if found_problem == False:
                message = "[ok]"
                process_status = True
    # check if a match for given file was found in log data
    if found_match == False:
        process_status = False
        message = "[Failed to retrieve processing data for file.]"
    return process_status, message


def get_file_name_from_data_block(message_block):
    """
    Extracts the file name from a process message block.

    :param message_block: list of json formatted rows representing all messages received during file process
    :type message_block: [str]

    :return: The fully qualified file path of the file processed
    :rtype: str
    """

    # check if this is a cloud model
    if "CLOUD MODEL" in message_block[0]:
        # ['\t- Processing file (1 of 1): CLOUD MODEL', '\t- ', '\t- \tProject ID: GUID', '\t- \tModel ID: GUID',...]
        message_starter = "\t- \t"
        file_name = message_block[3].Trim()[len(message_starter) - 1 :]
    else:
        # ["\t- Processing file (x of y): file path"}},...]
        message_starter = "\t- Processing file (x of y): "
        file_name = message_block[0].Trim()[len(message_starter) - 1 :]
    return file_name


def get_files_processed(file_path):
    """
    Reads a batch processor logfile and extracts all file names of files processed.

    :param file_path: The fully qualified file path to log file.
    :type file_path: str

    :return: a list lists containing The fully qualified file path for each file processed and their process status. True for no exception encountered, otherwise false.
        [[filepath, status]]
    :rtype: [[str, bool]]
    """
    # log file structure:
    # start of file list (file server):
    #   {"date":{"local":"17/12/2020","utc":"17/12/2020"},"time":{"local":"16:49:28","utc":"05:49:28"},"sessionId":"235e2180-dc33-4d61-8773-1005a59344c0","message":{"msgId":"","message":"Revit Files for processing (1):"}}
    #   {"date":{"local":"17/12/2020","utc":"17/12/2020"},"time":{"local":"16:49:28","utc":"05:49:28"},"sessionId":"235e2180-dc33-4d61-8773-1005a59344c0","message":{"msgId":"","message":""}}
    #   {"date":{"local":"17/12/2020","utc":"17/12/2020"},"time":{"local":"16:49:28","utc":"05:49:28"},"sessionId":"235e2180-dc33-4d61-8773-1005a59344c0","message":{"msgId":"","message":"\tP:\\something\\FileName.rvt"}}
    #   {"date":{"local":"17/12/2020","utc":"17/12/2020"},"time":{"local":"16:49:28","utc":"05:49:28"},"sessionId":"235e2180-dc33-4d61-8773-1005a59344c0","message":{"msgId":"","message":"\tFile exists: YES"}}
    #   {"date":{"local":"17/12/2020","utc":"17/12/2020"},"time":{"local":"16:49:28","utc":"05:49:28"},"sessionId":"235e2180-dc33-4d61-8773-1005a59344c0","message":{"msgId":"","message":"\tFile size: 86.93MB"}}
    #   {"date":{"local":"17/12/2020","utc":"17/12/2020"},"time":{"local":"16:49:28","utc":"05:49:28"},"sessionId":"235e2180-dc33-4d61-8773-1005a59344c0","message":{"msgId":"","message":"\tRevit version: Autodesk Revit 2020 (Build: 20200826_1250(x64))"}}
    #   {"date":{"local":"17/12/2020","utc":"17/12/2020"},"time":{"local":"16:49:28","utc":"05:49:28"},"sessionId":"235e2180-dc33-4d61-8773-1005a59344c0","message":{"msgId":"","message":""}}
    #   {"date":{"local":"17/12/2020","utc":"17/12/2020"},"time":{"local":"16:49:28","utc":"05:49:28"},"sessionId":"235e2180-dc33-4d61-8773-1005a59344c0","message":{"msgId":"","message":"Starting batch operation..."}}
    # end of file list
    # start of file list (cloud model):
    #   {"date":{"local":"18/05/2021","utc":"18/05/2021"},"time":{"local":"19:20:00","utc":"09:20:00"},"sessionId":"e8a39f45-ca88-464f-a32f-278f0414280f","message":{"msgId":"","message":"Revit Files for processing (1):"}}
    #   {"date":{"local":"18/05/2021","utc":"18/05/2021"},"time":{"local":"19:20:00","utc":"09:20:00"},"sessionId":"e8a39f45-ca88-464f-a32f-278f0414280f","message":{"msgId":"","message":""}}
    #   {"date":{"local":"18/05/2021","utc":"18/05/2021"},"time":{"local":"19:20:00","utc":"09:20:00"},"sessionId":"e8a39f45-ca88-464f-a32f-278f0414280f","message":{"msgId":"","message":"\tCLOUD MODEL"}}
    #   {"date":{"local":"18/05/2021","utc":"18/05/2021"},"time":{"local":"19:20:00","utc":"09:20:00"},"sessionId":"e8a39f45-ca88-464f-a32f-278f0414280f","message":{"msgId":"","message":"\tProject ID: ee514b99-fac1-441a-ba98-6912c4aa4fdb"}}
    #   {"date":{"local":"18/05/2021","utc":"18/05/2021"},"time":{"local":"19:20:00","utc":"09:20:00"},"sessionId":"e8a39f45-ca88-464f-a32f-278f0414280f","message":{"msgId":"","message":"\tModel ID: c15a664d-fab8-4153-8532-b8fa04402528"}}
    #   {"date":{"local":"18/05/2021","utc":"18/05/2021"},"time":{"local":"19:20:00","utc":"09:20:00"},"sessionId":"e8a39f45-ca88-464f-a32f-278f0414280f","message":{"msgId":"","message":"\tRevit version: 2020"}}
    #   {"date":{"local":"18/05/2021","utc":"18/05/2021"},"time":{"local":"19:20:00","utc":"09:20:00"},"sessionId":"e8a39f45-ca88-464f-a32f-278f0414280f","message":{"msgId":"","message":""}}
    # end of file list
    list_of_files = []
    json_data = read_log_file(file_path)
    # get data block showing which files are to be processed
    # there should just be one ...
    log_blocks = get_log_blocks(
        json_data, "Revit Files for processing", ["Starting batch operation..."], False
    )
    if len(log_blocks) > 0:
        unformatted_revit_file_process_messages = log_blocks[0]
        # parse data block and get list of files and file exists status
        # each file block is proceeded by an empty message row
        # last entry is also an empty message block!
        for x in range(len(unformatted_revit_file_process_messages)):
            # check for start of data block
            # Output('unformatted messages '+str(unformatted_revit_file_process_messages))
            if unformatted_revit_file_process_messages[x] == "" and x + 3 <= len(
                unformatted_revit_file_process_messages
            ):
                # check whether cloud model or file server model
                if "CLOUD MODEL" in unformatted_revit_file_process_messages[x + 1]:
                    # get file data from next two rows
                    # substitute file name with file GUID, fake the status always exists
                    dummy = [
                        unformatted_revit_file_process_messages[x + 3],
                        "File exists: YES",
                    ]
                else:
                    # get file data from next two rows
                    dummy = [
                        unformatted_revit_file_process_messages[x + 1],
                        unformatted_revit_file_process_messages[x + 2],
                    ]
                list_of_files.append(get_file_data(dummy))
    return list_of_files


# method parsing two rows of json formatted data
#
# data list of 2 rows of Json formatted data
# sample of file system:
# 18/05/2021 15:57:53 : 	File exists: YES
# 18/05/2021 15:57:53 : 	File size: 199.38MB
# sample of BIM360:
# 18/05/2021 15:37:23 : 	Project ID: a valid guid
# 18/05/2021 15:37:23 : 	Model ID: a valid guid
# note: when processing BIM360 files batch processor is not checking whether the file exists upfront!
# returns list in format
# [filename, file exists status as bool]
def get_file_data(data):
    # trim white spaces from file name
    file_name = data[0].Trim()
    file_status = False
    # check whether file status contains a YES or whether this is a cloud model
    # (RBP does not check upfront whether a cloud model exists!)
    if "YES" in data[1] or "Project ID" in data[1]:
        file_status = True
    return [file_name, file_status]


def get_files_not_found(files_processed):
    """
    Filters list of all files meant to be processed and returns the ones flagged as file not found.

    :param files_processed: array of lists in format: [[filename, status as bool],[filename, status as bool],...]
    :type files_processed: [[str, bool]]

    :return: array of lists in format: [[filename, status as bool],[filename, status as bool],...]
    :rtype: [[str, bool]]
    """

    file_not_found = []
    for f in files_processed:
        if f[1] == False:
            file_not_found.append(f)
    return file_not_found


def get_log_blocks(json_data, start_marker, end_markers, multiple_blocks):
    """
    Reads json formatted data into blocks per files processed.

    Returns the message sections per row entry only.

    :param json_data: List of logfile rows in json data format
    :type json_data: [str]
    :param start_marker: String in messages indicating start of block.
    :type start_marker: str
    :param end_markers: String in messages indicating end of block.
    :type end_markers: str
    :param multiple_blocks: Flag indicating whether there are multiple data block in log file to be returned
    :type multiple_blocks: bool

    :return: List of list of str
    :rtype: [[str]]
    """

    unformatted_block_data = []
    data_block = []
    message_string = ""
    # extract rows belonging to blocks
    file_block = False
    # Output('json data: ' + str(json_data))
    for data in json_data:
        message_string = get_message_from_json(data)
        if message_string.startswith(start_marker) and file_block == False:
            file_block = True
        # flag to indicate we found a match for end of block
        match_end_block = False
        for end_marker in end_markers:
            match = False
            if message_string.startswith(end_marker) and file_block == True:
                match_end_block = True
                unformatted_block_data.append(data_block)
                data_block = []
                file_block = False
                match = True
                break
        if not multiple_blocks and match:
            break
        if file_block:
            data_block.append(message_string)
    # check for open block
    if file_block == True and match_end_block == False:
        # append this data...hopefully there is an exception message in there!!
        unformatted_block_data.append(data_block)
        output("Added open data block")
        output(data_block)
        output("")
    return unformatted_block_data


def read_log_file(file_path):
    """
    Reads batch processor log file into lists of json objects.

    Sample row:
        {'sessionId': '778e87a5-4b94-4552-9e7e-c9ed38b5caee', 'time': {'local': '09:35:45', 'utc': '22:35:45'}, 'date': {'local': '25/11/2020', 'utc': '24/11/2020'}, 'message': {'msgId': '', 'message': ''}}

    :param file_path: The fully qualified file path to log file in json format
    :type file_path: str

    :return: list of json objects
    :rtype: [json]
    """

    data = []
    with open(file_path) as f:
        for line in f:
            data.append(json.loads(line))
    return data


def process_log_files(folder_path, debug=False):
    """
    Loops over log files and checks whether any exceptions occurred during revit files processing.

    :param folder_path: Fully qualified directory path where marker files are stored.
    :type folder_path: str
    :param debug: Flag indicating whether this is running in debug mode which will output some debug messages, defaults to False
    :type debug: bool, optional

    :return:
        Result class instance.

        - Process status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain message(s) for each revit file found in logs and whether an exception occurred during processing
        - result.result will be an empty list.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.
        - result.status will be an empty list.

    :rtype: :class:`.Result`

    """

    return_value = res.Result()
    debug_mode_ = debug
    logfile_results = []
    try:
        # get all marker files
        marker_file_ids = get_current_session_ids(folder_path)
        if debug_mode_:
            return_value.append_message(
                "Found marker file(s): {}".format(len(marker_file_ids))
            )
        if len(marker_file_ids) > 0:
            # find log files matching markers
            log_files = get_log_files(marker_file_ids)
            if debug_mode_:
                return_value.append_message(
                    "Found log file(s): {}".format(len(log_files))
                )
            if len(log_files) == len(marker_file_ids):
                data = []
                for lf in log_files:
                    # debug output
                    message = "Processing log file(s): {}".format(lf)
                    # return_value.AppendMessage('Processing log files: ' + lf)
                    try:
                        data = process_log_file(lf)
                        if len(data) > 0:
                            message = (
                                "{} [Got processed Revit file(s) data: {}]".format(
                                    message, len(data)
                                )
                            )
                        else:
                            # dummy run no files processed!
                            message = "{} [No Revit file(s) processed!]".format(message)
                    except Exception as e:
                        message = "{} [An exception occurred: {}]".format(message, e)
                    if debug_mode_:
                        return_value.append_message(message)
                    for d in data:
                        logfile_results.append(d)
                return_value.append_message("\n")
                return_value.append_message("Processed file(s) results:")
                # store results in return object
                for lf_results in logfile_results:
                    list_to_str = "\t".join(map(str, lf_results))
                    return_value.append_message(list_to_str)
                return_value.status = True
            else:
                return_value.update_sep(
                    False,
                    "Number of log files [{}] does not match required number: {}".format(
                        len(log_files), len(marker_file_ids)
                    ),
                )
        else:
            return_value.update_sep(
                False, "No marker files found in location: {}".format(folder_path)
            )
    except Exception as e:
        return_value.update_sep(False, "Terminated with Exception: {}".format(e))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Utility functions converting data retrieved from Revit into shapely geometry and processing it.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module requires python >3.9 due to dependencies:

- numpy
- shapely

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# these packages are not available in an iron python environment .e.g. Revit Python shell
# to avoid an exception stopping the entire package to load these are within a try catch block

import shapely.geometry as sg
import numpy as np

from duHast.Data.Objects import data_ceiling as dc
from duHast.Data.Objects import data_room as dr

from duHast.Data.Objects.Properties.Geometry import geometry_polygon

# --------------- shapely polygon creation ------------------


def get_translation_matrix(geometry_object):
    """
    Gets the rotation/ translation matrix from the geometry object

    :param geometry_object: A data geometry object instance.
    :type geometry_object: :class:`.DataGeometry`

    :return: A translation matrix.
    :rtype: numpy array
    """

    translation_matrix = []  # translation only matrix
    # note numpy creates arrays by row!
    # need to append one more row since matrix dot multiplication rule:
    # number of columns in first matrix must match number of rows in second matrix (point later on)
    for vector in geometry_object.rotation_coord:
        vector.append(0.0)
        translation_matrix.append(vector)
    rotation_matrix = geometry_object.translation_coord  # rotation matrix
    # adding extra row here
    rotation_matrix.append(1.0)
    translation_matrix.append(rotation_matrix)
    # build combined rotation and translation matrix
    combined_matrix = np.array(translation_matrix)
    # transpose matrix (translation matrix in json file is stored by columns not by rows!)
    combined_matrix = np.transpose(combined_matrix)
    return combined_matrix


def get_outer_loop_as_shapely_points(geometry_object, translation_matrix):
    """
    Returns the boundary loop of an object as list of shapely points.

    Points are translated with passed in matrix.
    Any loops containing less then 3 points will be ignored. (Empty list will be returned)

    :param geometry_object: A data geometry object instance.
    :type geometry_object: :class:`.DataGeometry`
    :param translation_matrix: A translation matrix.
    :type translation_matrix: numpy array

    :return: List of shapely points defining a polygon. (Empty list will be returned if less then 3 points in loop.)
    :rtype: List[shapely.point]
    """

    single_polygon_loop = []
    if geometry_object.data_type == geometry_polygon.DataPolygon.data_type:
        for point_double in geometry_object.outer_loop:
            # check if a 2D or 3D point
            if len(point_double) == 2:
                # assume a Z value of 0.0 (room loops are 2D)
                point_double.append(0.0)

            # need to add 1 to list for matrix multiplication
            # number of columns in first matrix (translation) must match number of rows in second matrix (point)
            translated_point = np.dot(
                translation_matrix,
                [point_double[0], point_double[1], point_double[2], 1.0],
            )
            p = sg.Point(translated_point[0], translated_point[1], translated_point[2])
            single_polygon_loop.append(p)
    # ignore any poly loops with less then 3 sides (less then 3 points)
    if len(single_polygon_loop) > 2:
        return single_polygon_loop
    else:
        return []


def get_inner_loops_as_shapely_points(geometry_object, translation_matrix):
    """
    Returns the inner loops (holes) of an object as list of lists of shapely points.

    Points are translated with passed in matrix.
    Any inner loops containing less then 3 points will be ignored. (Empty list will be returned)

    :param geoObject: A data geometry object instance.
    :type geoObject: :class:`.DataGeometry`
    :param translationM: A translation matrix.
    :type translationM: numpy array

    :return: List of lists of shapely points defining a polygon.
    :rtype: list [list[shapely.point]]
    """

    shapely_points = []
    # get inner loops
    if len(geometry_object.inner_loops) > 0:
        # there might be more then one inner loop
        for inner_loop in geometry_object.inner_loops:
            single_polygon_loop = []
            for point_double in inner_loop:
                # check if a 2D or 3D point
                if len(point_double) == 2:
                    # assume a Z value of 0.0 (room loops are 2D)
                    point_double.append(0.0)
                # need to add 1 to list for matrix multiplication
                # number of columns in first matrix (translation) must match number of rows in second matrix (point)
                translated_point = np.dot(
                    translation_matrix,
                    [point_double[0], point_double[1], point_double[2], 1.0],
                )
                p = sg.Point(
                    translated_point[0], translated_point[1], translated_point[2]
                )
                single_polygon_loop.append(p)
            # ignore any poly loops with less then 3 sides ( less then 3 points)
            if len(single_polygon_loop) > 2:
                shapely_points.append(single_polygon_loop)
    return shapely_points


def build_shapely_polygon(shapely_polygons):
    """
    Creates shapely polygon with nested polygons from list of shapely polygons past in.

    Assumptions is: first polygon describes the boundary loop and any subsequent polygons are describing\
         holes within the boundary 

    :param shapely_polygons: list of shapely polygons
    :type shapely_polygons: list[shapely.polygon]

    :return: A shapely polygon.
    :rtype: shapely.polygon
    """

    # convert to shapely
    poly = None
    # check if we got multiple polygons
    if len(shapely_polygons) == 1:
        # single exterior boundary ... no holes
        poly = sg.Polygon(shapely_polygons[0])
    elif len(shapely_polygons) > 1:
        # got holes...
        # set up interior holes to be added to polygon
        # (remember exterior point order is ccw, holes cw else
        # holes may not appear as holes.)
        interiors = {}
        for i in range(1, len(shapely_polygons)):
            interiors[i - 1] = shapely_polygons[i]
        i_p = {k: sg.Polygon(v) for k, v in interiors.items()}
        # create polygon with holes
        poly = sg.Polygon(
            shapely_polygons[0],
            [
                poly.exterior.coords
                for poly in i_p.values()
                if poly.within(sg.Polygon(shapely_polygons[0])) is True
            ],
        )
    return poly


def get_shapely_polygons_from_data_instance(data_instance):
    """
    Returns a list of of shapely polygons from data instances past in.

    Polygons may contain holes

    :param data_instance: _description_
    :type data_instance: A class with .geometry property returning a :class:`.DataGeometry` instance.

    :return: A list of shapely polygons.
    :rtype: list [shapely.polygon]
    """

    all_polygons = []
    # loop over data geometry and convert into shapely polygons

    for geometry_object in data_instance.polygon:
        if geometry_object.data_type == geometry_polygon.DataPolygon.data_type:
            translation_matrix = get_translation_matrix(geometry_object)
            shape_shapely = []
            outer_loop = get_outer_loop_as_shapely_points(
                geometry_object, translation_matrix
            )
            shape_shapely.append(outer_loop)
            if len(outer_loop) > 0:
                inner_loops = get_inner_loops_as_shapely_points(
                    geometry_object, translation_matrix
                )
                if len(inner_loops) > 0:
                    for l in inner_loops:
                        shape_shapely.append(l)
            poly = build_shapely_polygon(shape_shapely)
            all_polygons.append(poly)
        else:
            print("Not a polygon data instance: {}".format(geometry_object))
    return all_polygons


# --------------- end generics ------------------

#: List of available geometry (from revit to shapely ) converters
GEOMETRY_CONVERTER = {
    dr.DataRoom.data_type: get_shapely_polygons_from_data_instance,
    dc.DataCeiling.data_type: get_shapely_polygons_from_data_instance,
}


def get_shapely_polygons_from_geo_object(geometry_objects, data_type):
    """
    Converts polygon points from DataGeometry instances to shapely polygon instances and returns them as a dictionary where:

    - key is the geometry objects id
    - value is a list of shapely polygons

    :param geometry_objects: A list of instances of the the same type (i.e DataRoom)
    :type geometry_objects: list[data object]
    :param data_type: string human readable identifying the data type ( each Data... class has this as a static field: dr.DataRoom.dataType)
    :type data_type: str

    :return: A dictionary.
    :rtype: {int:[shapely.polygon]}
    """

    multi_polygons = {}
    for i in range(len(geometry_objects)):
        multi_polygons[geometry_objects[i].instance_properties.id] = []
        poly = GEOMETRY_CONVERTER[data_type](geometry_objects[i])
        for p in poly:
            if p != None:
                multi_polygons[geometry_objects[i].instance_properties.id].append(p)
    return multi_polygons



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit points helper functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import (
    Transform,
    UV,
    XYZ,
)
from duHast.Utilities.compare import is_close


def rotate_point_around_z_with_origin(point, origin, angle_in_radians):
    """
    Rotates a point around the Z-axis with a given origin and angle in radians.
    Returns the rotated point.

    :param point: The point to rotate.
    :type point: Autodesk.Revit.DB.XYZ
    :param origin: The origin of the rotation.
    :type origin: Autodesk.Revit.DB.XYZ
    :param angle_in_radians: The angle in radians to rotate the point.
    :type angle_in_radians: float

    :return: The rotated point.
    :rtype: Autodesk.Revit.DB.XYZ
    """

    # return point unchanged if angle is 0
    if angle_in_radians == 0:
        return point

    # get the axis of rotation
    rotation_axis = XYZ.BasisZ
    # Create a rotation transform around the Z-axis
    rotation_transform = Transform.CreateRotationAtPoint(
        rotation_axis, angle_in_radians, origin
    )

    # Apply the rotation transform to the point
    rotated_point = rotation_transform.OfPoint(point)

    return rotated_point


def transform_point_by_elem_transform(pt, transform):
    """
    Transforms a point by an element transform
    :param pt: The point to transform
    :type pt: XYZ
    :param transform: The transform to use
    :type transform: Transform
    :return: The transformed point
    :rtype: XYZ
    """

    x = pt.X
    y = pt.Y
    z = pt.Z

    b0 = transform.get_Basis(0)
    b1 = transform.get_Basis(1)
    b2 = transform.get_Basis(2)
    origin = transform.Origin

    x_new = x * b0.X + y * b1.X + z * b2.X + origin.X
    y_new = x * b0.Y + y * b1.Y + z * b2.Y + origin.Y
    z_new = x * b0.Z + y * b1.Z + z * b2.Z + origin.Z

    return XYZ(x_new, y_new, z_new)


def sort_points_by_min_and_max(min_pt, max_pt):
    """
    Takes BoundingBox or Outline Min and Max points and
    returns the true Min and Max points. This is to ensure
    no zero thickness geometries are created.
    :param min_pt: The minimum point
    :type min_pt: XYZ
    :param max_pt: The maximum point
    :type max_pt: XYZ
    :return: The sorted points
    :rtype: tuple
    """
    min_x = min_pt.X
    min_y = min_pt.Y
    min_z = min_pt.Z

    max_x = max_pt.X
    max_y = max_pt.Y
    max_z = max_pt.Z

    smin_x = min(min_x, max_x)
    smin_y = min(min_y, max_y)
    smax_x = max(min_x, max_x)
    smax_y = max(min_y, max_y)

    return (XYZ(smin_x, smin_y, min_z), XYZ(smax_x, smax_y, max_z))


def get_point_as_doubles(point, include_z=False):
    """
    Converts a revit XYZ to a list of doubles in order x,y,z.

    :param point: A revit point.
    :type point: Autodesk.Revit.DB.XYZ

    :return: List of doubles in order of x,y,z
    :rtype: list double
    """
    if include_z:
        return [point.X, point.Y, point.Z]
    else:
        return [point.X, point.Y]


def get_doubles_as_xyz(doubles, include_z=False):
    """
    Converts a list of doubles to a revit XYZ.

    :param doubles: List of doubles in order of x,y,z
    :type doubles: list double

    :return: A revit point.
    :rtype: Autodesk.Revit.DB.XYZ
    """
    if include_z:
        return XYZ(doubles[0], doubles[1], doubles[2])
    else:
        return XYZ(doubles[0], doubles[1], 0)


def get_rotation_around_z_as_xyz(doubles):
    """
    Converts a list of doubles to a revit XYZ.
    The first two doubles are the x,y coordinates of the point.
    The third double is the z coordinate of the point to which 1 is added.

    :param doubles: List of doubles in order of x,y,z
    :type doubles: list double

    :return: A revit point.
    :rtype: Autodesk.Revit.DB.XYZ
    """

    return XYZ(doubles[0], doubles[1], doubles[2] + 1)


def flatten_xyz_point(point):
    """
    Flattens a XYZ point to a UV by omitting the Z value of the XYZ.

    https://thebuildingcoder.typepad.com/blog/2008/12/2d-polygon-areas-and-outer-loop.html

    :param point: A revit point.
    :type point: Autodesk.Revit.DB.XYZ

    :return: A 2D point (UV)
    :rtype:  Autodesk.Revit.DB.UV
    """

    return UV(point.X, point.Y)


def are_points_identical(p1, p2):
    """
    Compares the X,Y,Z values of two revit point and returns True if they are the same, otherwise False

    :param p1: A revit point.
    :type p1: Autodesk.Revit.DB.XYZ
    :param p2: A revit point.
    :type p2: Autodesk.Revit.DB.XYZ

    :return: True if they are the same, otherwise False.
    :rtype: bool
    """

    if is_close(p1.X, p2.X) and is_close(p1.Y, p2.Y) and is_close(p1.Z, p2.Z):
        return True
    else:
        return False


def check_duplicate_point(points, point):
    """
    Checks whether a collection of points contains another given point and returns True if that is the case.

    :param points: List of revit points
    :type points: list Autodesk.Revit.DB.XYZ
    :param point: A revit point.
    :type point: Autodesk.Revit.DB.XYZ

    :return: True if point is in collection, otherwise False.
    :rtype: bool
    """

    for p1 in points:
        if are_points_identical(p1, point):
            return True
    return False


# ---------------------------- debug ----------------------------
def get_point_as_string(point):
    """
    Returns Revit point as a string.

    :param point: A revit point.
    :type point: Autodesk.Revit.DB.XYZ

    :return: String in format 'X:Y:Z'
    :rtype: str
    """

    return str(point.X) + " : " + str(point.Y) + " : " + str(point.Z)



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Customizable element filter actions which can be used with the custom element filter class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Thw two action in this module check whether:

- a given property does contain / match a set of test value
- a given property does not contain / match a set of test value


actions expect:

- the test to be performed
- the test values to be performed against
- an output function to pipe any exception message to



"""


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


def action_element_property_contains_any_of_values(test_values, test, output):
    """

    Set up a function checking whether element name contains any of the test values.

    Note:
    If element names does contain any of the test values, the function will return True, Otherwise False.

    :param values: List of values the element name may contain.
    :type values: [str]
    :param output: A function piping the string output to a required location.(console...)
    :type output: func(message)

    """

    def action(doc, element_id):
        flag = False
        try:
            element = doc.GetElement(element_id)
            for value in test_values:
                if test(value, element):
                    flag = True
                    break
        except Exception as e:
            output(
                "action_element_property_contains_values terminated with exception: {}".format(
                    e
                )
            )
            flag = False
        return flag

    return action


def action_element_property_does_not_contains_any_of_values(test_values, test, output):
    """

    Set up a function checking whether element name does not contains any of the test values.

    Note:
    If element names does contain any of the test values, the function will return False, Otherwise True.

    :param values: List of values the element name may not contain.
    :type values: [str]
    :param output: A function piping the string output to a required location.(console...)
    :type output: func(message)
    """

    def action(doc, element_id):
        flag = True
        try:
            element = doc.GetElement(element_id)
            for value in test_values:
                if test(value, element):
                    flag = False
                    break
        except Exception as e:
            output(
                "action_element_property_does_ not_contains_values terminated with exception: {}".format(
                    e
                )
            )
            flag = False
        return flag

    return action



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A class to store file information.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# An item to represent a file name in a row in a grid.

from duHast.Utilities.Objects import base


class MyFileItem(base.Base):
    def __init__(
        self,
        name,
        size,
        bim360_project_guid=None,
        bim360_file_guid=None,
        bim360_revit_version="-",
        is_selected = False,
    ):
        """
        Class constructor.

        :param name: The fully qualified file path.
        :type name: str
        :param size: The file size.
        :type size: int
        :param  bim360_project_guid: The BIM360 project GUID, defaults to None
        :type  bim360_project_guid: str, optional
        :param bim360_file_guid: The BIM360 file GUID, defaults to None
        :type bim360_file_guid: str, optional
        :param bim360_revit_version: The revit file version (year only), defaults to '-'
        :type bim360_revit_version: str, optional
        """

        # ini super class to allow multi inheritance in children!
        super(MyFileItem, self).__init__()

        self.name = name
        self.size = size
        self.bim_360_project_guid = bim360_project_guid
        self.bim_360_file_guid = bim360_file_guid
        self.bim_360_revit_version = bim360_revit_version
        self.is_selected = is_selected

    def __eq__(self, other):
        """
        Check if two objects are equal.

        :param other: The other object to compare to.
        :type other: object
        :return: True if the objects are equal, False otherwise.
        :rtype: bool
        """

        if isinstance(other, MyFileItem) and (
            self.name,
            self.size,
            self.bim_360_project_guid ,
            self.bim_360_file_guid ,
            self.bim_360_revit_version ,
            
        ) == (
            other.name,
            other.size,
            other.bim_360_project_guid ,
            other.bim_360_file_guid ,
            other.bim_360_revit_version ,
        ):
            return True
        else:
            return False
    
    # python 2.7 needs custom implementation of not equal
    def __ne__(self, other):
        return not self.__eq__(other=other)
        


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Purge action storage class for Revit purge unused.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


# import System
# import clr

from duHast.Utilities.Objects import base


class PurgeAction(base.Base):
    def __init__(
        self,
        purgeTransactionName,  # the function which returns all element ids to be purged
        purgeIdsGetter,  # the function which returns all element ids to be purged
        purgeReportHeader,  # human readable report header for each test action
        testReportHeader,  # human readable report header for each test action
        testIdsGetter,  # functions which returns all available type ids in model of same category as purge action. To be used to compare ids before and after coded purge with ids before and after revit built in purge
    ):
        """
        Class constructor.

        :param purgeTransactionName: The name of the transaction to purge elements under.
        :type purgeTransactionName: str
        :param purgeIdsGetter: The function which returns all element ids to be purged
        :type purgeIdsGetter: func
        :param purgeReportHeader: Human readable report header for each purge action
        :type purgeReportHeader: str
        :param testReportHeader: Human readable report header for each test action
        :type testReportHeader: str
        :param testIdsGetter: functions which returns all available type ids in model of same category as purge action.\
            To be used to compare ids before and after code purge with ids before and after revit built in purge
        :type testIdsGetter: func
        """

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(PurgeAction, self).__init__()

        self.purge_transaction_name = purgeTransactionName
        self.purge_ids_getter = purgeIdsGetter
        self.purge_report_header = purgeReportHeader
        self.test_report_header = testReportHeader
        self.test_ids_getter = testIdsGetter
